# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import random
import socket
import time
from multiprocessing import Process

import merlin
import pytest
import requests
from merlin.model import ModelType

request_json = {"instances": [[2.8, 1.0, 6.8, 0.4], [3.1, 1.4, 4.5, 1.6]]}

if os.environ.get("CI_SERVER"):
    host = "172.17.0.1"
else:
    host = "localhost"


@pytest.mark.integration
@pytest.mark.local_server_test
@pytest.mark.dependency(
    depends=["test/integration_test.py::test_sklearn"], scope="session"
)
def test_sklearn(integration_test_url, project_name, use_google_oauth):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("sklearn-sample", ModelType.SKLEARN)
    v = _get_latest_version(merlin.active_model())
    port = _get_free_port()
    p = Process(target=v.start_server, kwargs={"port": port, "build_image": True})
    p.start()
    _wait_server_ready(f"http://{host}:{port}")
    resp = requests.post(_get_local_endpoint(v, port), json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])
    p.terminate()


@pytest.mark.integration
@pytest.mark.local_server_test
@pytest.mark.dependency(
    depends=["test/integration_test.py::test_xgboost"], scope="session"
)
def test_xgboost(integration_test_url, project_name, use_google_oauth):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("xgboost-sample", ModelType.XGBOOST)
    v = _get_latest_version(merlin.active_model())
    port = _get_free_port()
    p = Process(target=v.start_server, kwargs={"port": port, "build_image": True})
    p.start()
    _wait_server_ready(f"http://{host}:{port}")
    resp = requests.post(_get_local_endpoint(v, port), json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])
    p.terminate()


@pytest.mark.integration
@pytest.mark.local_server_test
@pytest.mark.dependency(
    depends=["test/integration_test.py::test_tensorflow"], scope="session"
)
def test_tensorflow(integration_test_url, project_name, use_google_oauth):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("tensorflow-sample", ModelType.TENSORFLOW)
    v = _get_latest_version(merlin.active_model())
    port = _get_free_port()
    p = Process(target=v.start_server, kwargs={"port": port, "build_image": True})
    p.start()
    _wait_server_ready(f"http://{host}:{port}/v1/models/{v.model.name}-{v.id}")
    request_json = {
        "signature_name": "predict",
        "instances": [
            {
                "sepal_length": 2.8,
                "sepal_width": 1.0,
                "petal_length": 6.8,
                "petal_width": 0.4,
            },
            {
                "sepal_length": 0.1,
                "sepal_width": 0.5,
                "petal_length": 1.8,
                "petal_width": 2.4,
            },
        ],
    }
    resp = requests.post(_get_local_endpoint(v, port), json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])
    p.terminate()


@pytest.mark.integration
@pytest.mark.local_server_test
@pytest.mark.dependency(
    depends=["test/integration_test.py::test_pytorch"], scope="session"
)
def test_pytorch(integration_test_url, project_name, use_google_oauth):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pytorch-sample", ModelType.PYTORCH)
    model_dir = "test/pytorch-model"

    with merlin.new_model_version() as v:
        merlin.log_model(model_dir=model_dir)

    port = _get_free_port()
    p = Process(target=v.start_server, kwargs={"port": port, "build_image": True})
    p.start()
    _wait_server_ready(f"http://{host}:{port}")
    resp = requests.post(_get_local_endpoint(v, port), json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])
    p.terminate()


@pytest.mark.local_server_test
@pytest.mark.integration
@pytest.mark.dependency(
    depends=["test/pyfunc_integration_test.py::test_pyfunc"], scope="session"
)
def test_pyfunc(integration_test_url, project_name, use_google_oauth):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pyfunc-sample", ModelType.PYFUNC)
    v = _get_latest_version(merlin.active_model())
    port = _get_free_port()
    p = Process(target=v.start_server, kwargs={"port": port})
    p.start()
    _wait_server_ready(f"http://{host}:{port}", timeout_second=600)
    resp = requests.post(_get_local_endpoint(v, port), json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])
    p.terminate()


def _get_latest_version(model):
    v_list = model.list_version()
    v_list.sort(key=lambda v: v.id, reverse=True)
    return v_list[0]


def _get_free_port():
    sock = None
    try:
        while True:
            port = random.randint(8000, 9000)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = sock.connect_ex(("127.0.0.1", port))
            if result != 0:
                return port
    finally:
        if sock is not None:
            sock.close()


def _wait_server_ready(url, timeout_second=300, tick_second=5):
    ellapsed_second = 0
    while ellapsed_second < timeout_second:
        try:
            resp = requests.get(url)
            if resp.status_code == 200:
                return
        except Exception as e:
            print(f"{url} is not ready: {e}")

        time.sleep(tick_second)
        ellapsed_second += tick_second

    if ellapsed_second >= timeout_second:
        raise TimeoutError("server is not ready within specified timeout duration")


def _get_local_endpoint(model_version, port):
    return f"http://{host}:{port}/v1/models/{model_version.model.name}-{model_version.id}:predict"



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictBool, StrictInt, StrictStr
from client.models.gpu_config import GPUConfig
from client.models.prediction_job_resource_request import PredictionJobResourceRequest
from client.models.resource_request import ResourceRequest
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class Environment(BaseModel):
    """
    Environment
    """ # noqa: E501
    id: Optional[StrictInt] = None
    name: StrictStr
    cluster: StrictStr
    is_default: Optional[StrictBool] = None
    region: Optional[StrictStr] = None
    gcp_project: Optional[StrictStr] = None
    default_resource_request: Optional[ResourceRequest] = None
    default_transformer_resource_request: Optional[ResourceRequest] = None
    default_prediction_job_resource_request: Optional[PredictionJobResourceRequest] = None
    gpus: Optional[List[GPUConfig]] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    __properties: ClassVar[List[str]] = ["id", "name", "cluster", "is_default", "region", "gcp_project", "default_resource_request", "default_transformer_resource_request", "default_prediction_job_resource_request", "gpus", "created_at", "updated_at"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of Environment from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of default_resource_request
        if self.default_resource_request:
            _dict['default_resource_request'] = self.default_resource_request.to_dict()
        # override the default output from pydantic by calling `to_dict()` of default_transformer_resource_request
        if self.default_transformer_resource_request:
            _dict['default_transformer_resource_request'] = self.default_transformer_resource_request.to_dict()
        # override the default output from pydantic by calling `to_dict()` of default_prediction_job_resource_request
        if self.default_prediction_job_resource_request:
            _dict['default_prediction_job_resource_request'] = self.default_prediction_job_resource_request.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in gpus (list)
        _items = []
        if self.gpus:
            for _item in self.gpus:
                if _item:
                    _items.append(_item.to_dict())
            _dict['gpus'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of Environment from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "id": obj.get("id"),
            "name": obj.get("name"),
            "cluster": obj.get("cluster"),
            "is_default": obj.get("is_default"),
            "region": obj.get("region"),
            "gcp_project": obj.get("gcp_project"),
            "default_resource_request": ResourceRequest.from_dict(obj.get("default_resource_request")) if obj.get("default_resource_request") is not None else None,
            "default_transformer_resource_request": ResourceRequest.from_dict(obj.get("default_transformer_resource_request")) if obj.get("default_transformer_resource_request") is not None else None,
            "default_prediction_job_resource_request": PredictionJobResourceRequest.from_dict(obj.get("default_prediction_job_resource_request")) if obj.get("default_prediction_job_resource_request") is not None else None,
            "gpus": [GPUConfig.from_dict(_item) for _item in obj.get("gpus")] if obj.get("gpus") is not None else None,
            "created_at": obj.get("created_at"),
            "updated_at": obj.get("updated_at")
        })
        return _obj





# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictInt, StrictStr
from client.models.config import Config
from client.models.environment import Environment
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class PredictionJob(BaseModel):
    """
    PredictionJob
    """ # noqa: E501
    id: Optional[StrictInt] = None
    name: Optional[StrictStr] = None
    version_id: Optional[StrictInt] = None
    model_id: Optional[StrictInt] = None
    project_id: Optional[StrictInt] = None
    environment_name: Optional[StrictStr] = None
    environment: Optional[Environment] = None
    config: Optional[Config] = None
    status: Optional[StrictStr] = None
    error: Optional[StrictStr] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    __properties: ClassVar[List[str]] = ["id", "name", "version_id", "model_id", "project_id", "environment_name", "environment", "config", "status", "error", "created_at", "updated_at"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of PredictionJob from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of environment
        if self.environment:
            _dict['environment'] = self.environment.to_dict()
        # override the default output from pydantic by calling `to_dict()` of config
        if self.config:
            _dict['config'] = self.config.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of PredictionJob from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "id": obj.get("id"),
            "name": obj.get("name"),
            "version_id": obj.get("version_id"),
            "model_id": obj.get("model_id"),
            "project_id": obj.get("project_id"),
            "environment_name": obj.get("environment_name"),
            "environment": Environment.from_dict(obj.get("environment")) if obj.get("environment") is not None else None,
            "config": Config.from_dict(obj.get("config")) if obj.get("config") is not None else None,
            "status": obj.get("status"),
            "error": obj.get("error"),
            "created_at": obj.get("created_at"),
            "updated_at": obj.get("updated_at")
        })
        return _obj





# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import types
from unittest.mock import patch

import client
import client as cl
import pytest
from merlin import AutoscalingPolicy, DeploymentMode, MetricsType
from merlin.autoscaling import (
    RAW_DEPLOYMENT_DEFAULT_AUTOSCALING_POLICY,
    SERVERLESS_DEFAULT_AUTOSCALING_POLICY,
)
from merlin.batch.config import PredictionJobConfig, ResultType
from merlin.batch.job import JobStatus
from merlin.batch.sink import BigQuerySink, SaveMode
from merlin.batch.source import BigQuerySource
from merlin.endpoint import VersionEndpoint
from merlin.model import ModelType
from merlin.model_schema import InferenceSchema, ModelSchema, RankingOutput, ValueType
from merlin.protocol import Protocol
from urllib3_mock import Responses

responses = Responses("requests.packages.urllib3")

default_resource_request = cl.ResourceRequest(
    min_replica=1, max_replica=1, cpu_request="100m", memory_request="128Mi"
)
gpu = cl.GPUConfig(
    name="nvidia-tesla-p4",
    values=["1", "4", "8"],
    resource_type="nvidia.com/gpu",
    node_selector={"cloud.google.com/gke-accelerator": "nvidia-tesla-p4"},
    tolerations=[
        cl.GPUToleration(
            key="caraml/nvidia-tesla-p4",
            operator="Equal",
            value="enabled",
            effect="NoSchedule",
        ),
        cl.GPUToleration(
            key="nvidia.com/gpu", operator="Equal", value="present", effect="NoSchedule"
        ),
    ],
)

env_1 = cl.Environment(
    id=1,
    name="dev",
    cluster="cluster-1",
    is_default=True,
    default_resource_request=default_resource_request,
)
env_2 = cl.Environment(
    id=2,
    name="dev-2",
    cluster="cluster-2",
    is_default=False,
    default_resource_request=default_resource_request,
)
env_3 = cl.Environment(
    id=2,
    name="dev-3",
    cluster="cluster-3",
    is_default=False,
    default_resource_request=default_resource_request,
    gpus=[gpu],
)

ep1 = cl.VersionEndpoint(
    id="1234",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_1.name,
    environment=env_1,
    monitoring_url="grafana.com",
)
ep2 = cl.VersionEndpoint(
    id="4567",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-2",
    environment_name=env_2.name,
    environment=env_2,
    monitoring_url="grafana.com",
)
ep3 = cl.VersionEndpoint(
    id="1234",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_1.name,
    environment=env_1,
    monitoring_url="grafana.com",
    deployment_mode="raw_deployment",
)
ep4 = cl.VersionEndpoint(
    id="1234",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_1.name,
    environment=env_1,
    monitoring_url="grafana.com",
    autoscaling_policy=client.AutoscalingPolicy(
        metrics_type=cl.MetricsType.CPU_UTILIZATION, target_value=10
    ),
)
resource_request_with_gpu = cl.ResourceRequest(
    min_replica=1,
    max_replica=1,
    cpu_request="100m",
    memory_request="128Mi",
    gpu_name="nvidia-tesla-p4",
    gpu_request="1",
)
ep5 = cl.VersionEndpoint(
    id="789",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_3.name,
    environment=env_3,
    monitoring_url="grafana.com",
    resource_request=resource_request_with_gpu,
)
upi_ep = cl.VersionEndpoint(
    id="1234",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_1.name,
    environment=env_1,
    monitoring_url="grafana.com",
    protocol=cl.Protocol.UPI_V1,
)
observability_enabled_ep = cl.VersionEndpoint(
    id="7899",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_3.name,
    environment=env_3,
    monitoring_url="grafana.com",
    enable_model_observability=True,
)

rule_1 = cl.ModelEndpointRule(
    destinations=[
        cl.ModelEndpointRuleDestination(version_endpoint_id=ep1.id, weight=100)
    ]
)
rule_2 = cl.ModelEndpointRule(
    destinations=[
        cl.ModelEndpointRuleDestination(version_endpoint_id=ep2.id, weight=100)
    ]
)
mdl_endpoint_1 = cl.ModelEndpoint(
    id=1,
    model_id=1,
    model=None,
    status="serving",
    url="localhost/1",
    rule=rule_1,
    environment_name=env_1.name,
    environment=env_1,
)
mdl_endpoint_2 = cl.ModelEndpoint(
    id=2,
    model_id=1,
    model=None,
    status="serving",
    url="localhost/2",
    rule=rule_2,
    environment_name=env_2.name,
    environment=env_2,
)

mdl_endpoint_upi = cl.ModelEndpoint(
    id=1,
    model_id=1,
    model=None,
    status="serving",
    url="localhost/1",
    rule=rule_1,
    environment_name=env_1.name,
    environment=env_1,
    protocol=cl.Protocol.UPI_V1,
)

config = {
    "job_config": {
        "version": "v1",
        "kind": "PredictionJob",
        "name": "job-1",
        "bigquery_source": {
            "table": "project.dataset.source_table",
            "features": ["feature_1", "feature_2"],
            "options": {
                "key": "val",
            },
        },
        "model": {
            "type": "PYFUNC_V2",
            "uri": "gs://my-model/model",
            "result": {"type": "DOUBLE"},
            "options": {
                "key": "val",
            },
        },
        "bigquery_sink": {
            "table": "project.dataset.result_table",
            "staging_bucket": "gs://test",
            "result_column": "prediction",
            "save_mode": 1,
            "options": {
                "key": "val",
            },
        },
    },
    "image_ref": "asia.gcr.io/my-image:1",
    "service_account_name": "my-service-account",
    "resource_request": {
        "driver_cpu_request": "1",
        "driver_memory_request": "1Gi",
        "executor_cpu_request": "1",
        "executor_memory_request": "1Gi",
        "executor_replica": 1,
    },
}
job_1 = cl.PredictionJob(
    id=1,
    name="job-1",
    version_id=1,
    model_id=1,
    config=config,
    status="pending",
    error="",
    created_at="2019-08-29T08:13:12.377Z",
    updated_at="2019-08-29T08:13:12.377Z",
)
job_2 = cl.PredictionJob(
    id=2,
    name="job-2",
    version_id=1,
    model_id=1,
    config=config,
    status="pending",
    error="error",
    created_at="2019-08-29T08:13:12.377Z",
    updated_at="2019-08-29T08:13:12.377Z",
)


def serialize_datetime(obj):
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()
    raise TypeError("Type is not serializable")


class TestProject:
    secret_1 = cl.Secret(id=1, name="secret-1", data="secret-data-1")
    secret_2 = cl.Secret(id=2, name="secret-2", data="secret-data-2")

    @responses.activate
    def test_create_secret(self, project):
        responses.add(
            "POST",
            "/v1/projects/1/secrets",
            body=json.dumps(self.secret_1.to_dict()),
            status=200,
            content_type="application/json",
        )

        project.create_secret(self.secret_1.name, self.secret_1.data)
        actual_body = json.loads(responses.calls[0].request.body)
        assert actual_body["name"] == self.secret_1.name
        assert actual_body["data"] == self.secret_1.data

    @responses.activate
    def test_update_secret(self, project):
        responses.add(
            "GET",
            "/v1/projects/1/secrets",
            body=json.dumps([self.secret_1.to_dict(), self.secret_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "PATCH",
            "/v1/projects/1/secrets/1",
            body=json.dumps(self.secret_1.to_dict()),
            status=200,
            content_type="application/json",
        )

        project.update_secret(self.secret_1.name, "new-data")

        actual_body = json.loads(responses.calls[1].request.body)
        assert actual_body["name"] == self.secret_1.name
        assert actual_body["data"] == "new-data"

        responses.reset()

        # test secret not found
        responses.add(
            "GET",
            "/v1/projects/1/secrets",
            body=json.dumps([self.secret_1.to_dict()]),
            status=200,
            content_type="application/json",
        )

        with pytest.raises(
            ValueError,
            match=f"unable to find secret {self.secret_2.name} in project {project.name}",
        ):
            project.update_secret(self.secret_2.name, "new-data")

    @responses.activate
    def test_delete_secret(self, project):
        responses.add(
            "GET",
            "/v1/projects/1/secrets",
            body=json.dumps([self.secret_1.to_dict(), self.secret_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "DELETE",
            "/v1/projects/1/secrets/1",
            status=204,
            content_type="application/json",
        )

        project.delete_secret(self.secret_1.name)

        responses.reset()

        # test secret not found
        responses.add(
            "GET",
            "/v1/projects/1/secrets",
            body=json.dumps([self.secret_1.to_dict()]),
            status=200,
            content_type="application/json",
        )

        with pytest.raises(
            ValueError,
            match=f"unable to find secret {self.secret_2.name} in project {project.name}",
        ):
            project.delete_secret(self.secret_2.name)

    @responses.activate
    def test_list_secret(self, project):
        responses.add(
            "GET",
            "/v1/projects/1/secrets",
            body=json.dumps([self.secret_1.to_dict(), self.secret_2.to_dict()]),
            status=200,
            content_type="application/json",
        )

        secret_names = project.list_secret()
        assert secret_names == [self.secret_1.name, self.secret_2.name]


class TestModelVersion:
    @responses.activate
    def test_list_endpoint(self, version):
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep1.to_dict(), ep2.to_dict()]),
            status=200,
            content_type="application/json",
        )

        endpoints = version.list_endpoint()
        assert len(endpoints) == 2
        assert endpoints[0].id == ep1.id
        assert endpoints[1].id == ep2.id

    @responses.activate
    def test_deploy(self, version):
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        # This is the additional check which deploy makes to determine if there are any existing endpoints associated
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "POST",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps(ep1.to_dict()),
            status=201,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint/1234",
            body=json.dumps(ep1.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep1.to_dict()]),
            status=200,
            content_type="application/json",
        )

        endpoint = version.deploy(environment_name=env_1.name)

        assert endpoint.id == ep1.id
        assert endpoint.status.value == ep1.status
        assert endpoint.environment_name == ep1.environment_name
        assert endpoint.environment.cluster == env_1.cluster
        assert endpoint.environment.name == env_1.name
        assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
        assert endpoint.autoscaling_policy == SERVERLESS_DEFAULT_AUTOSCALING_POLICY
        assert endpoint.protocol == Protocol.HTTP_JSON

    @responses.activate
    def test_deploy_upiv1(self, version):
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        # This is the additional check which deploy makes to determine if there are any existing endpoints associated
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "POST",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps(upi_ep.to_dict()),
            status=201,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint/1234",
            body=json.dumps(upi_ep.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([upi_ep.to_dict()]),
            status=200,
            content_type="application/json",
        )

        endpoint = version.deploy(environment_name=env_1.name)

        assert endpoint.id == upi_ep.id
        assert endpoint.status.value == upi_ep.status
        assert endpoint.environment_name == upi_ep.environment_name
        assert endpoint.environment.cluster == env_1.cluster
        assert endpoint.environment.name == env_1.name
        assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
        assert endpoint.autoscaling_policy == SERVERLESS_DEFAULT_AUTOSCALING_POLICY
        assert endpoint.protocol == Protocol.UPI_V1

    @responses.activate
    def test_deploy_using_raw_deployment_mode(self, version):
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        # This is the additional check which deploy makes to determine if there are any existing endpoints associated
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "POST",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps(ep3.to_dict()),
            status=201,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint/1234",
            body=json.dumps(ep3.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep3.to_dict()]),
            status=200,
            content_type="application/json",
        )

        endpoint = version.deploy(
            environment_name=env_1.name, deployment_mode=DeploymentMode.RAW_DEPLOYMENT
        )

        assert endpoint.id == ep3.id
        assert endpoint.status.value == ep3.status
        assert endpoint.environment_name == ep3.environment_name
        assert endpoint.environment.cluster == env_1.cluster
        assert endpoint.environment.name == env_1.name
        assert endpoint.deployment_mode == DeploymentMode.RAW_DEPLOYMENT
        assert endpoint.autoscaling_policy == RAW_DEPLOYMENT_DEFAULT_AUTOSCALING_POLICY

    @responses.activate
    def test_deploy_with_autoscaling_policy(self, version):
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        # This is the additional check which deploy makes to determine if there are any existing endpoints associated
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "POST",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps(ep4.to_dict()),
            status=201,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint/1234",
            body=json.dumps(ep4.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep4.to_dict()]),
            status=200,
            content_type="application/json",
        )

        endpoint = version.deploy(
            environment_name=env_1.name,
            autoscaling_policy=AutoscalingPolicy(
                metrics_type=MetricsType.CPU_UTILIZATION, target_value=10
            ),
        )

        assert endpoint.id == ep4.id
        assert endpoint.status.value == ep4.status
        assert endpoint.environment_name == ep4.environment_name
        assert endpoint.environment.cluster == env_1.cluster
        assert endpoint.environment.name == env_1.name
        assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
        assert endpoint.autoscaling_policy.metrics_type == MetricsType.CPU_UTILIZATION
        assert endpoint.autoscaling_policy.target_value == 10

    @responses.activate
    def test_deploy_default_env(self, version):
        # This is the additional check which deploy makes to determine if there are any existing endpoints associated
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        # no default environment
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        with pytest.raises(ValueError):
            version.deploy()

        # default environment exists
        responses.reset()
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        # This is the additional check which deploy makes to determine if there are any existing endpoints associated
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "POST",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps(ep1.to_dict()),
            status=201,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint/1234",
            body=json.dumps(ep1.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep1.to_dict()]),
            status=200,
            content_type="application/json",
        )

        endpoint = version.deploy()

        assert endpoint.id == ep1.id
        assert endpoint.status.value == ep1.status
        assert endpoint.environment_name == ep1.environment_name
        assert endpoint.environment.cluster == env_1.cluster
        assert endpoint.environment.name == env_1.name

    @responses.activate
    def test_redeploy_model(self, version):
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        # This is the additional check which deploy makes to determine if there are any existing endpoints associated
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep3.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "PUT",
            "/v1/models/1/versions/1/endpoint/1234",
            body=json.dumps(ep4.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint/1234",
            body=json.dumps(ep4.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep4.to_dict()]),
            status=200,
            content_type="application/json",
        )

        # Redeployment (add autoscaling policy and change deployment mode)
        endpoint = version.deploy(
            environment_name=env_1.name,
            autoscaling_policy=AutoscalingPolicy(
                metrics_type=MetricsType.CPU_UTILIZATION, target_value=10
            ),
        )

        assert endpoint.id == ep4.id
        assert endpoint.status.value == ep4.status
        assert endpoint.environment_name == ep4.environment_name
        assert endpoint.environment.cluster == env_1.cluster
        assert endpoint.environment.name == env_1.name
        assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
        assert endpoint.autoscaling_policy.metrics_type == MetricsType.CPU_UTILIZATION
        assert endpoint.autoscaling_policy.target_value == 10

    @responses.activate
    def test_deploy_with_gpu(self, version):
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_3.to_dict()]),
            status=200,
            content_type="application/json",
        )
        # This is the additional check which deploy makes to determine if there are any existing endpoints associated
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "POST",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps(ep5.to_dict()),
            status=201,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint/789",
            body=json.dumps(ep5.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep5.to_dict()]),
            status=200,
            content_type="application/json",
        )

        endpoint = version.deploy(environment_name=env_3.name)

        assert endpoint.id == ep5.id
        assert endpoint.status.value == ep5.status
        assert endpoint.environment_name == ep5.environment_name
        assert endpoint.environment.cluster == env_3.cluster
        assert endpoint.environment.name == env_3.name
        assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
        assert endpoint.resource_request.gpu_name == resource_request_with_gpu.gpu_name
        assert (
            endpoint.resource_request.gpu_request
            == resource_request_with_gpu.gpu_request
        )

    @responses.activate
    def test_deploy_with_model_observability_enabled(self, version):
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_3.to_dict()]),
            status=200,
            content_type="application/json",
        )
        # This is the additional check which deploy makes to determine if there are any existing endpoints associated
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "POST",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps(observability_enabled_ep.to_dict()),
            status=201,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([observability_enabled_ep.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint/7899",
            body=json.dumps(observability_enabled_ep.to_dict()),
            status=200,
            content_type="application/json",
        )

        endpoint = version.deploy(
            environment_name=env_3.name, enable_model_observability=True
        )

        assert endpoint.id == observability_enabled_ep.id
        assert endpoint.status.value == observability_enabled_ep.status
        assert endpoint.environment_name == observability_enabled_ep.environment_name
        assert endpoint.environment.cluster == env_3.cluster
        assert endpoint.environment.name == env_3.name
        assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
        assert endpoint.enable_model_observability == True

    @responses.activate
    def test_undeploy(self, version):
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep2.to_dict()]),
            status=200,
            content_type="application/json",
        )

        version.undeploy(environment_name=env_1.name)
        assert len(responses.calls) == 1

        responses.reset()
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep1.to_dict(), ep2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "DELETE",
            "/v1/models/1/versions/1/endpoint/1234",
            body=json.dumps(ep1.to_dict()),
            status=200,
            content_type="application/json",
        )

        version.undeploy(environment_name=env_1.name)
        assert len(responses.calls) == 2

    @responses.activate
    def test_undeploy_default_env(self, version):
        # This is the additional check which deploy makes to determine if there are any existing endpoints associated
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        # no default environment
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        with pytest.raises(ValueError):
            version.deploy()

        responses.reset()
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )

        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep2.to_dict()]),
            status=200,
            content_type="application/json",
        )

        version.undeploy()
        assert len(responses.calls) == 2

        responses.reset()
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions/1/endpoint",
            body=json.dumps([ep1.to_dict(), ep2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "DELETE",
            "/v1/models/1/versions/1/endpoint/1234",
            body=json.dumps(ep1.to_dict()),
            status=200,
            content_type="application/json",
        )

        version.undeploy()
        assert len(responses.calls) == 3

    @responses.activate
    def test_list_prediction_job(self, version):
        responses.add(
            method="GET",
            url="/v1/models/1/versions/1/jobs-by-page?page=1",
            body=json.dumps({
                "results": [job_1.to_dict()],
                "paging": {
                    "page": 1,
                    "pages": 2,
                    "total": 2,
                },
            }, default=serialize_datetime),
            status=200,
            content_type="application/json",
            match_querystring=True,
        )
        responses.add(
            method="GET",
            url="/v1/models/1/versions/1/jobs-by-page?page=2",
            body=json.dumps({
                "results": [job_2.to_dict()],
                "paging": {
                    "page": 2,
                    "pages": 2,
                    "total": 2,
                },
            }, default=serialize_datetime),
            status=200,
            content_type="application/json",
            match_querystring=True,
        )
        jobs = version.list_prediction_job()
        assert len(jobs) == 2
        assert jobs[0].id == job_1.id
        assert jobs[0].name == job_1.name
        assert jobs[0].status == JobStatus(job_1.status)
        assert jobs[0].error == job_1.error

        assert jobs[1].id == job_2.id
        assert jobs[1].name == job_2.name
        assert jobs[1].status == JobStatus(job_2.status)
        assert jobs[1].error == job_2.error

    @responses.activate
    def test_create_prediction_job(self, version):
        job_1.status = "completed"
        responses.add(
            "POST",
            "/v1/models/1/versions/1/jobs",
            body=json.dumps(job_1.to_dict(), default=serialize_datetime),
            status=200,
            content_type="application/json",
        )

        bq_src = BigQuerySource(
            table="project.dataset.source_table",
            features=["feature_1", "feature2"],
            options={"key": "val"},
        )

        bq_sink = BigQuerySink(
            table="project.dataset.result_table",
            result_column="prediction",
            save_mode=SaveMode.OVERWRITE,
            staging_bucket="gs://test",
            options={"key": "val"},
        )

        job_config = PredictionJobConfig(
            source=bq_src,
            sink=bq_sink,
            service_account_name="my-service-account",
            result_type=ResultType.INTEGER,
        )

        j = version.create_prediction_job(job_config=job_config)
        assert j.status == JobStatus.COMPLETED
        assert j.id == job_1.id
        assert j.error == job_1.error
        assert j.name == job_1.name

        actual_req = json.loads(responses.calls[0].request.body)
        assert actual_req["config"]["job_config"]["bigquery_source"] == bq_src.to_dict()
        assert actual_req["config"]["job_config"]["bigquery_sink"] == bq_sink.to_dict()
        assert (
            actual_req["config"]["job_config"]["model"]["result"]["type"]
            == ResultType.INTEGER.value
        )
        assert (
            actual_req["config"]["job_config"]["model"]["uri"]
            == f"{version.artifact_uri}/model"
        )
        assert (
            actual_req["config"]["job_config"]["model"]["type"]
            == ModelType.PYFUNC_V2.value.upper()
        )
        assert actual_req["config"]["service_account_name"] == "my-service-account"

    @patch("merlin.model.DEFAULT_PREDICTION_JOB_DELAY", 0)
    @patch("merlin.model.DEFAULT_PREDICTION_JOB_RETRY_DELAY", 0)
    @responses.activate
    def test_create_prediction_job_with_retry_failed(self, version):
        job_1.status = "pending"
        responses.add(
            "POST",
            "/v1/models/1/versions/1/jobs",
            body=json.dumps(job_1.to_dict(), default=serialize_datetime),
            status=200,
            content_type="application/json",
        )

        for i in range(5):
            responses.add(
                "GET",
                "/v1/models/1/versions/1/jobs/1",
                body=json.dumps(job_1.to_dict(), default=serialize_datetime),
                status=500,
                content_type="application/json",
            )

        bq_src = BigQuerySource(
            table="project.dataset.source_table",
            features=["feature_1", "feature2"],
            options={"key": "val"},
        )

        bq_sink = BigQuerySink(
            table="project.dataset.result_table",
            result_column="prediction",
            save_mode=SaveMode.OVERWRITE,
            staging_bucket="gs://test",
            options={"key": "val"},
        )

        job_config = PredictionJobConfig(
            source=bq_src,
            sink=bq_sink,
            service_account_name="my-service-account",
            result_type=ResultType.INTEGER,
        )

        with pytest.raises(ValueError):
            j = version.create_prediction_job(job_config=job_config)
            assert j.id == job_1.id
            assert j.error == job_1.error
            assert j.name == job_1.name
            assert len(responses.calls) == 6

    @patch("merlin.model.DEFAULT_PREDICTION_JOB_DELAY", 0)
    @patch("merlin.model.DEFAULT_PREDICTION_JOB_RETRY_DELAY", 0)
    @responses.activate
    def test_create_prediction_job_with_retry_success(self, version):
        job_1.status = "pending"
        responses.add(
            "POST",
            "/v1/models/1/versions/1/jobs",
            body=json.dumps(job_1.to_dict(), default=serialize_datetime),
            status=200,
            content_type="application/json",
        )

        # Patch the method as currently it is not supported in the library
        # https://github.com/getsentry/responses/issues/135
        def _find_match(self, request):
            for match in self._urls:
                if request.method == match["method"] and self._has_url_match(
                    match, request.url
                ):
                    return match

        def _find_match_patched(self, request):
            for index, match in enumerate(self._urls):
                if request.method == match["method"] and self._has_url_match(
                    match, request.url
                ):
                    if (
                        request.method == "GET"
                        and request.url == "/v1/models/1/versions/1/jobs/1"
                    ):
                        return self._urls.pop(index)
                    else:
                        return match

        responses._find_match = types.MethodType(_find_match_patched, responses)

        for i in range(4):
            responses.add(
                "GET",
                "/v1/models/1/versions/1/jobs/1",
                body=json.dumps(job_1.to_dict(), default=serialize_datetime),
                status=500,
                content_type="application/json",
            )

        job_1.status = "completed"
        responses.add(
            "GET",
            "/v1/models/1/versions/1/jobs/1",
            body=json.dumps(job_1.to_dict(), default=serialize_datetime),
            status=200,
            content_type="application/json",
        )

        bq_src = BigQuerySource(
            table="project.dataset.source_table",
            features=["feature_1", "feature2"],
            options={"key": "val"},
        )

        bq_sink = BigQuerySink(
            table="project.dataset.result_table",
            result_column="prediction",
            save_mode=SaveMode.OVERWRITE,
            staging_bucket="gs://test",
            options={"key": "val"},
        )

        job_config = PredictionJobConfig(
            source=bq_src,
            sink=bq_sink,
            service_account_name="my-service-account",
            result_type=ResultType.INTEGER,
        )

        j = version.create_prediction_job(job_config=job_config)
        assert j.status == JobStatus.COMPLETED
        assert j.id == job_1.id
        assert j.error == job_1.error
        assert j.name == job_1.name

        actual_req = json.loads(responses.calls[0].request.body)
        assert actual_req["config"]["job_config"]["bigquery_source"] == bq_src.to_dict()
        assert actual_req["config"]["job_config"]["bigquery_sink"] == bq_sink.to_dict()
        assert (
            actual_req["config"]["job_config"]["model"]["result"]["type"]
            == ResultType.INTEGER.value
        )
        assert (
            actual_req["config"]["job_config"]["model"]["uri"]
            == f"{version.artifact_uri}/model"
        )
        assert (
            actual_req["config"]["job_config"]["model"]["type"]
            == ModelType.PYFUNC_V2.value.upper()
        )
        assert actual_req["config"]["service_account_name"] == "my-service-account"
        assert len(responses.calls) == 6

        # unpatch
        responses._find_match = types.MethodType(_find_match, responses)

    @patch("merlin.model.DEFAULT_PREDICTION_JOB_DELAY", 0)
    @patch("merlin.model.DEFAULT_PREDICTION_JOB_RETRY_DELAY", 0)
    @responses.activate
    def test_create_prediction_job_with_retry_pending_then_failed(self, version):
        job_1.status = "pending"
        responses.add(
            "POST",
            "/v1/models/1/versions/1/jobs",
            body=json.dumps(job_1.to_dict(), default=serialize_datetime),
            status=200,
            content_type="application/json",
        )

        # Patch the method as currently it is not supported in the library
        # https://github.com/getsentry/responses/issues/135
        def _find_match(self, request):
            for match in self._urls:
                if request.method == match["method"] and self._has_url_match(
                    match, request.url
                ):
                    return match

        def _find_match_patched(self, request):
            for index, match in enumerate(self._urls):
                if request.method == match["method"] and self._has_url_match(
                    match, request.url
                ):
                    if (
                        request.method == "GET"
                        and request.url == "/v1/models/1/versions/1/jobs/1"
                    ):
                        return self._urls.pop(index)
                    else:
                        return match

        responses._find_match = types.MethodType(_find_match_patched, responses)

        for i in range(3):
            responses.add(
                "GET",
                "/v1/models/1/versions/1/jobs/1",
                body=json.dumps(job_1.to_dict(), default=serialize_datetime),
                status=500,
                content_type="application/json",
            )

        responses.add(
            "GET",
            "/v1/models/1/versions/1/jobs/1",
            body=json.dumps(job_1.to_dict(), default=serialize_datetime),
            status=200,
            content_type="application/json",
        )

        job_1.status = "failed"
        for i in range(5):
            responses.add(
                "GET",
                "/v1/models/1/versions/1/jobs/1",
                body=json.dumps(job_1.to_dict(), default=serialize_datetime),
                status=500,
                content_type="application/json",
            )

        bq_src = BigQuerySource(
            table="project.dataset.source_table",
            features=["feature_1", "feature2"],
            options={"key": "val"},
        )

        bq_sink = BigQuerySink(
            table="project.dataset.result_table",
            result_column="prediction",
            save_mode=SaveMode.OVERWRITE,
            staging_bucket="gs://test",
            options={"key": "val"},
        )

        job_config = PredictionJobConfig(
            source=bq_src,
            sink=bq_sink,
            service_account_name="my-service-account",
            result_type=ResultType.INTEGER,
        )

        with pytest.raises(ValueError):
            j = version.create_prediction_job(job_config=job_config)
            assert j.id == job_1.id
            assert j.error == job_1.error
            assert j.name == job_1.name

        # unpatch
        responses._find_match = types.MethodType(_find_match, responses)
        assert len(responses.calls) == 10

    @responses.activate
    def test_stop_prediction_job(self, version):
        job_1.status = "pending"
        responses.add(
            "POST",
            "/v1/models/1/versions/1/jobs",
            body=json.dumps(job_1.to_dict(), default=serialize_datetime),
            status=200,
            content_type="application/json",
        )

        responses.add(
            "PUT",
            "/v1/models/1/versions/1/jobs/1/stop",
            status=204,
            content_type="application/json",
        )

        job_1.status = "terminated"
        responses.add(
            "GET",
            "/v1/models/1/versions/1/jobs/1",
            body=json.dumps(job_1.to_dict(), default=serialize_datetime),
            status=200,
            content_type="application/json",
        )

        bq_src = BigQuerySource(
            table="project.dataset.source_table",
            features=["feature_1", "feature2"],
            options={"key": "val"},
        )

        bq_sink = BigQuerySink(
            table="project.dataset.result_table",
            result_column="prediction",
            save_mode=SaveMode.OVERWRITE,
            staging_bucket="gs://test",
            options={"key": "val"},
        )

        job_config = PredictionJobConfig(
            source=bq_src,
            sink=bq_sink,
            service_account_name="my-service-account",
            result_type=ResultType.INTEGER,
        )

        j = version.create_prediction_job(job_config=job_config, sync=False)
        j = j.stop()
        assert j.status == JobStatus.TERMINATED
        assert j.id == job_1.id
        assert j.error == job_1.error
        assert j.name == job_1.name

    @responses.activate
    def test_model_version_deletion(self, version):
        responses.add(
            "DELETE",
            "/v1/models/1/versions/1",
            body=json.dumps(1),
            status=200,
            content_type="application/json",
        )
        response = version.delete_model_version()

        assert response == 1


class TestModel:
    v1 = cl.Version(id=1, model_id=1)
    v2 = cl.Version(id=2, model_id=1)
    v3 = cl.Version(
        id=3, model_id=1, labels={"model": "T-800"}, python_version="3.10.*"
    )

    schema = client.ModelSchema(
        id=3,
        model_id=1,
        spec=client.SchemaSpec(
            tag_columns=["tags", "extras"],
            feature_types={
                "featureA": client.ValueType.FLOAT64,
                "featureB": client.ValueType.INT64,
                "featureC": client.ValueType.BOOLEAN,
                "featureD": client.ValueType.STRING,
            },
            model_prediction_output=client.ModelPredictionOutput(
                client.RankingOutput(
                    rank_score_column="score",
                    relevance_score_column="relevance_score",
                    output_class=client.ModelPredictionOutputClass.RANKINGOUTPUT,
                )
            ),
        ),
    )
    merlin_model_schema = ModelSchema(
        id=3,
        model_id=1,
        spec=InferenceSchema(
            tag_columns=["tags", "extras"],
            feature_types={
                "featureA": ValueType.FLOAT64,
                "featureB": ValueType.INT64,
                "featureC": ValueType.BOOLEAN,
                "featureD": ValueType.STRING,
            },
            model_prediction_output=RankingOutput(
                rank_score_column="score", relevance_score_column="relevance_score"
            ),
        ),
    )
    v4 = cl.Version(
        id=4,
        model_id=1,
        labels={"model": "T-800"},
        python_version="3.10.*",
        model_schema=schema,
    )

    @responses.activate
    def test_list_version(self, model):
        responses.add(
            "GET",
            "/v1/models/1/versions?limit=50&cursor=&search=",
            match_querystring=True,
            body=json.dumps([self.v1.to_dict()]),
            status=200,
            adding_headers={"Next-Cursor": "abcdef"},
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/versions?limit=50&cursor=abcdef&search=",
            match_querystring=True,
            body=json.dumps([self.v2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        versions = model.list_version()
        assert len(versions) == 2
        assert versions[0].id == 1
        assert versions[1].id == 2

    @responses.activate
    def test_list_version_with_labels(self, model):
        responses.add(
            "GET",
            "/v1/models/1/versions?limit=50&cursor=&search=labels%3Amodel+in+%28T-800%29",
            body=json.dumps([self.v3.to_dict()]),
            status=200,
            match_querystring=True,
            content_type="application/json",
        )
        versions = model.list_version({"model": ["T-800"]})
        assert len(versions) == 1
        assert versions[0].id == 3
        assert versions[0].labels["model"] == "T-800"

    @responses.activate
    def test_list_endpoint(self, model):
        responses.add(
            "GET",
            "/v1/models/1/endpoints",
            body=json.dumps([mdl_endpoint_1.to_dict(), mdl_endpoint_2.to_dict()]),
            status=200,
            content_type="application/json",
        )

        endpoints = model.list_endpoint()
        assert len(endpoints) == 2
        assert endpoints[0].id == mdl_endpoint_1.id
        assert endpoints[1].id == mdl_endpoint_2.id

    @responses.activate
    def test_new_model_version(self, model):
        responses.add(
            "POST",
            "/v1/models/1/versions",
            body=json.dumps(self.v4.to_dict()),
            status=201,
            content_type="application/json",
        )

        mv = model.new_model_version(
            labels={"model": "T-800"}, model_schema=self.merlin_model_schema
        )
        assert mv._python_version == "3.10.*"
        assert mv._id == 4
        assert mv._model._id == 1
        assert mv._labels == {"model": "T-800"}
        assert mv._model_schema == self.merlin_model_schema

    @responses.activate
    def test_serve_traffic(self, model):
        ve = VersionEndpoint(ep1)
        with pytest.raises(ValueError):
            model.serve_traffic([ve], environment_name=env_1.name)

        with pytest.raises(ValueError):
            model.serve_traffic({ve: -1}, environment_name=env_1.name)

        with pytest.raises(ValueError):
            model.serve_traffic({ve: 101}, environment_name=env_1.name)

        with pytest.raises(ValueError):
            model.serve_traffic(
                {VersionEndpoint(ep2): 100}, environment_name=env_1.name
            )

        # test create
        responses.add(
            "GET",
            "/v1/models/1/endpoints",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "POST",
            "/v1/models/1/endpoints",
            body=json.dumps(mdl_endpoint_1.to_dict()),
            status=201,
            content_type="application/json",
        )
        endpoint = model.serve_traffic({ve: 100}, environment_name=env_1.name)
        assert endpoint.id == mdl_endpoint_1.id
        assert (
            endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
        )
        assert endpoint.protocol == Protocol.HTTP_JSON

        responses.reset()

        # test update
        responses.add(
            "GET",
            "/v1/models/1/endpoints",
            body=json.dumps([mdl_endpoint_1.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/endpoints/1",
            body=json.dumps(mdl_endpoint_1.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "PUT",
            "/v1/models/1/endpoints/1",
            body=json.dumps(mdl_endpoint_1.to_dict()),
            status=200,
            content_type="application/json",
        )
        endpoint = model.serve_traffic({ve: 100}, environment_name=env_1.name)
        assert endpoint.id == mdl_endpoint_1.id
        assert (
            endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
        )

    @responses.activate
    def test_stop_serving_traffic(self, model):
        ve = VersionEndpoint(ep1)
        with pytest.raises(ValueError):
            model.serve_traffic([ve], environment_name=env_1.name)

        with pytest.raises(ValueError):
            model.serve_traffic({ve: -1}, environment_name=env_1.name)

        with pytest.raises(ValueError):
            model.serve_traffic({ve: 101}, environment_name=env_1.name)

        with pytest.raises(ValueError):
            model.serve_traffic(
                {VersionEndpoint(ep2): 100}, environment_name=env_1.name
            )

        # test create
        responses.add(
            "GET",
            "/v1/models/1/endpoints",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "POST",
            "/v1/models/1/endpoints",
            body=json.dumps(mdl_endpoint_1.to_dict()),
            status=201,
            content_type="application/json",
        )
        endpoint = model.serve_traffic({ve: 100}, environment_name=env_1.name)
        assert endpoint.id == mdl_endpoint_1.id
        assert (
            endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
        )

        responses.reset()

        # test DELETE
        responses.reset()
        responses.add(
            "GET",
            "/v1/models/1/endpoints",
            body=json.dumps([mdl_endpoint_1.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/endpoints/1",
            body=json.dumps(mdl_endpoint_1.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "DELETE",
            "/v1/models/1/endpoints/1",
            status=200,
            content_type="application/json",
        )
        model.stop_serving_traffic(endpoint.environment_name)
        assert len(responses.calls) == 2

    @responses.activate
    def test_serve_traffic_default_env(self, model):
        ve = VersionEndpoint(ep1)

        # no default environment
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        with pytest.raises(ValueError):
            model.serve_traffic({ve: 100})

        responses.reset()

        # test create
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/endpoints",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "POST",
            "/v1/models/1/endpoints",
            body=json.dumps(mdl_endpoint_1.to_dict()),
            status=201,
            content_type="application/json",
        )
        endpoint = model.serve_traffic({ve: 100})
        assert endpoint.id == mdl_endpoint_1.id
        assert (
            endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
        )

        responses.reset()

        # test update
        responses.add(
            "GET",
            "/v1/environments",
            body=json.dumps([env_1.to_dict(), env_2.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/endpoints",
            body=json.dumps([mdl_endpoint_1.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/endpoints/1",
            body=json.dumps(mdl_endpoint_1.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "PUT",
            "/v1/models/1/endpoints/1",
            body=json.dumps(mdl_endpoint_1.to_dict()),
            status=200,
            content_type="application/json",
        )
        endpoint = model.serve_traffic({ve: 100})
        assert endpoint.id == mdl_endpoint_1.id
        assert (
            endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
        )

    @responses.activate
    def test_serve_traffic_upi(self, model):
        ve = VersionEndpoint(upi_ep)
        # test create
        responses.add(
            "GET",
            "/v1/models/1/endpoints",
            body=json.dumps([]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "POST",
            "/v1/models/1/endpoints",
            body=json.dumps(mdl_endpoint_upi.to_dict()),
            status=201,
            content_type="application/json",
        )
        endpoint = model.serve_traffic({ve: 100}, environment_name=env_1.name)
        assert endpoint.id == mdl_endpoint_upi.id
        assert (
            endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
        )
        assert endpoint.protocol == Protocol.UPI_V1

        responses.reset()

        # test update
        responses.add(
            "GET",
            "/v1/models/1/endpoints",
            body=json.dumps([mdl_endpoint_upi.to_dict()]),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "GET",
            "/v1/models/1/endpoints/1",
            body=json.dumps(mdl_endpoint_upi.to_dict()),
            status=200,
            content_type="application/json",
        )
        responses.add(
            "PUT",
            "/v1/models/1/endpoints/1",
            body=json.dumps(mdl_endpoint_upi.to_dict()),
            status=200,
            content_type="application/json",
        )
        endpoint = model.serve_traffic({ve: 100}, environment_name=env_1.name)
        assert endpoint.id == mdl_endpoint_upi.id
        assert (
            endpoint.environment_name == env_1.name == mdl_endpoint_upi.environment_name
        )
        assert endpoint.protocol == Protocol.UPI_V1

    @responses.activate
    def test_model_deletion(self, model):
        responses.add(
            "DELETE",
            "/v1/projects/1/models/1",
            body=json.dumps(1),
            status=200,
            content_type="application/json",
        )
        response = model.delete_model()

        assert response == 1



import abc
from dataclasses import dataclass, field
from enum import Enum, unique
from typing import Dict, List, Optional

import numpy as np
import pandas as pd
from dataclasses_json import DataClassJsonMixin, config, dataclass_json


DEFAULT_SESSION_ID_COLUMN = "session_id"
DEFAULT_ROW_ID_COLUMN = "row_id"


class ObservationType(Enum):
    """
    Supported observation types.
    """

    FEATURE = "feature"
    PREDICTION = "prediction"
    GROUND_TRUTH = "ground_truth"


@unique
class ValueType(Enum):
    """
    Supported feature value types.
    """

    FLOAT64 = "float64"
    INT64 = "int64"
    BOOLEAN = "boolean"
    STRING = "string"


class PredictionOutput(abc.ABC):
    subclass_registry: dict = {}
    discriminator_field: str = "output_class"

    """
    Register subclasses of PredictionOutput.
    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        PredictionOutput.subclass_registry[cls.__name__] = cls

    """
    Given a subclass of PredictionOutput, which is assumed to have dataclass json mix in, encode
    the object with a discriminator field used to differentiate the different subclasses.
    """

    @classmethod
    def encode_with_discriminator(cls, x):
        if type(x) not in cls.subclass_registry.values():
            raise ValueError(
                f"Input must be a subclass of {cls.__name__}, got {type(x)}"
            )
        if not isinstance(x, DataClassJsonMixin):
            raise ValueError(
                f"Input must be a virtual subclass of DataClassJsonMixin, got {type(x)}"
            )

        return {
            cls.discriminator_field: type(x).__name__,
            **x.to_dict(),
        }

    """
    Given a dictionary, encoded using :func:`merlin.observability.inference.PredictionOutput.encode_with_discriminator`,
    decode the dictionary back to the correct subclass of PredictionOutput.
    """

    @classmethod
    def decode(cls, input: Dict):
        return PredictionOutput.subclass_registry[
            input[cls.discriminator_field]
        ].from_dict(input)

    """
    Given an input dataframe, return a new dataframe with the necessary columns for observability,
    along with a schema for the observability backend to parse the dataframe. In place changes might
    be made to the input dataframe.
    
    :param df: Input dataframe.
    :param session_id_column: Name of the column containing the session id.
    :param row_id_column: Name of the column containing the row id.
    :param observation_types: Types of observations to be included in the output dataframe.
    :return: output dataframe
    """

    @abc.abstractmethod
    def preprocess(
        self, df: pd.DataFrame, session_id_column: str, row_id_column: str, observation_types: List[ObservationType]
    ) -> pd.DataFrame:
        raise NotImplementedError

    """
    Return a dictionary mapping the name of the prediction output column to its value type.
    """

    @abc.abstractmethod
    def prediction_types(self) -> Dict[str, ValueType]:
        raise NotImplementedError

    """
    Return a dictionary mapping the name of the ground truth output column to its value type.
    """
    @abc.abstractmethod
    def ground_truth_types(self) -> Dict[str, ValueType]:
        raise NotImplementedError


@dataclass_json
@dataclass
class RegressionOutput(PredictionOutput):
    """
    Regression model prediction output schema.
    Attributes:
        prediction_score_column: Name of the column containing the prediction score.
        actual_score_column: Name of the column containing the actual score.
    """

    prediction_score_column: str
    actual_score_column: str

    def preprocess(
        self, df: pd.DataFrame, session_id_column: str, row_id_column: str, observation_types: List[ObservationType]
    ) -> pd.DataFrame:
        return df

    def prediction_types(self) -> Dict[str, ValueType]:
        return {
            self.prediction_score_column: ValueType.FLOAT64,
        }

    def ground_truth_types(self) -> Dict[str, ValueType]:
        return {
            self.actual_score_column: ValueType.FLOAT64,
        }


@dataclass_json
@dataclass
class BinaryClassificationOutput(PredictionOutput):
    """
    Binary classification model prediction output schema.

    Attributes:
        prediction_score_column: Name of the column containing the prediction score.
            Prediction score must be between 0.0 and 1.0.
        actual_score_column: Name of the column containing the actual score.
        positive_class_label: Label for positive class.
        negative_class_label: Label for negative class.
        score_threshold: Score threshold for prediction to be considered as positive class.
    """

    prediction_score_column: str
    actual_score_column: str
    positive_class_label: str
    negative_class_label: str
    score_threshold: float = 0.5

    @property
    def prediction_label_column(self) -> str:
        return "_prediction_label"

    @property
    def actual_label_column(self) -> str:
        return "_actual_label"

    def prediction_label(self, prediction_score: float) -> str:
        """
        Returns either positive or negative class label based on prediction score.
        :param prediction_score: Probability of positive class, between 0.0 and 1.0.
        :return: prediction label
        """
        return (
            self.positive_class_label
            if prediction_score >= self.score_threshold
            else self.negative_class_label
        )

    def actual_label(self, actual_score: int) -> str:
        """
        Derive actual label from actual score.
        :param actual_score: Actual score.
        :return: actual label
        """
        return (
            self.positive_class_label
            if actual_score >= self.score_threshold
            else self.negative_class_label
        )

    def preprocess(
        self, df: pd.DataFrame, session_id_column: str, row_id_column: str, observation_types: List[ObservationType]
    ) -> pd.DataFrame:
        if ObservationType.PREDICTION in observation_types:
            df[self.prediction_label_column] = df[self.prediction_score_column].apply(
                self.prediction_label
            )
        if ObservationType.GROUND_TRUTH in observation_types:
            df[self.actual_label_column] = df[self.actual_score_column].apply(
                self.actual_label
            )
        return df

    def prediction_types(self) -> Dict[str, ValueType]:
        return {
            self.prediction_score_column: ValueType.FLOAT64,
            self.prediction_label_column: ValueType.STRING
        }

    def ground_truth_types(self) -> Dict[str, ValueType]:
        return {
            self.actual_score_column: ValueType.FLOAT64,
            self.actual_label_column: ValueType.STRING
        }


@dataclass_json
@dataclass
class RankingOutput(PredictionOutput):
    rank_score_column: str
    relevance_score_column: str
    """
    Ranking model prediction output schema.
    Attributes:
        rank_score_column: Name of the column containing the ranking score of the prediction.
        relevance_score_column: Name of the column containing the relevance score of the prediction.
    """

    @property
    def rank_column(self) -> str:
        return "_rank"

    def preprocess(
        self, df: pd.DataFrame, session_id_column: str, row_id_column: str, observation_types: List[ObservationType]
    ) -> pd.DataFrame:
        if ObservationType.PREDICTION in observation_types:
            df[self.rank_column] = (
                df.groupby(session_id_column)[self.rank_score_column]
                .rank(method="first", ascending=False)
                .astype(np.int_)
            )
        return df

    def prediction_types(self) -> Dict[str, ValueType]:
        return {
            self.rank_score_column: ValueType.FLOAT64,
            self.rank_column: ValueType.INT64,
        }

    def ground_truth_types(self) -> Dict[str, ValueType]:
        return {
            self.relevance_score_column: ValueType.FLOAT64,
        }


@dataclass_json
@dataclass
class InferenceSchema:
    feature_types: Dict[str, ValueType]
    model_prediction_output: PredictionOutput = field(
        metadata=config(
            encoder=PredictionOutput.encode_with_discriminator,
            decoder=PredictionOutput.decode,
        )
    )
    session_id_column: str = DEFAULT_SESSION_ID_COLUMN
    row_id_column: str = DEFAULT_ROW_ID_COLUMN
    feature_orders: Optional[List[str]] = None
    tag_columns: Optional[List[str]] = None

    @property
    def feature_columns(self) -> List[str]:
        return list(self.feature_types.keys())

    @property
    def prediction_columns(self) -> List[str]:
        return list(self.model_prediction_output.prediction_types().keys())

    @property
    def ground_truth_columns(self) -> List[str]:
        return list(self.model_prediction_output.ground_truth_types().keys())

    def preprocess(
        self, df: pd.DataFrame, observation_types: List[ObservationType]
    ) -> pd.DataFrame:
        return self.model_prediction_output.preprocess(
            df, self.session_id_column, self.row_id_column, observation_types
        )

def add_prediction_id_column(df: pd.DataFrame, session_id_column: str, row_id_column: str, prediction_id_column: str) -> pd.DataFrame:
    """
    Add prediction id column to the dataframe.
    :param df: Input dataframe.
    :param session_id_column: Name of the column containing the session id.
    :param row_id_column: Name of the column containing the row id.
    :param prediction_id_column: Name of the prediction id column.
    :return: Dataframe with prediction id column.
    """
    df[prediction_id_column] = df[session_id_column] + "_" + df[row_id_column]
    return df


# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
from test.utils import undeploy_all_version
from time import sleep

import pandas as pd
import pytest
from merlin.logger import Logger, LoggerConfig, LoggerMode
from merlin.model import ModelType
from merlin.resource_request import ResourceRequest
from merlin.transformer import StandardTransformer, Transformer
from merlin.model_schema import ModelSchema
from merlin.observability.inference import InferenceSchema, ValueType, BinaryClassificationOutput
from recursive_diff import recursive_eq

import merlin
from merlin import DeploymentMode, MetricsType

request_json = {"instances": [[2.8, 1.0, 6.8, 0.4], [3.1, 1.4, 4.5, 1.6]]}
tensorflow_request_json = {
    "signature_name": "predict",
    "instances": [
        {
            "sepal_length": 2.8,
            "sepal_width": 1.0,
            "petal_length": 6.8,
            "petal_width": 0.4,
        },
        {
            "sepal_length": 0.1,
            "sepal_width": 0.5,
            "petal_length": 1.8,
            "petal_width": 2.4,
        },
    ],
}


@pytest.mark.integration
@pytest.mark.dependency()
def test_model_version_with_labels(
    integration_test_url, project_name, use_google_oauth
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("sklearn-labels", ModelType.SKLEARN)

    model_dir = "test/sklearn-model"

    undeploy_all_version()

    with merlin.new_model_version(labels={"model": "T-800"}) as v:
        # Upload the serialized model to MLP
        merlin.log_model(model_dir=model_dir)
        assert len(v.labels) == 1
        assert v.labels["model"] == "T-800"

    merlin_active_model = merlin.active_model()
    all_versions = merlin_active_model.list_version(labels={"model": ["T-800"]})
    for version in all_versions:
        assert version.labels["model"] == "T-800"

    should_not_exist_versions = merlin_active_model.list_version(
        labels={"model": ["T-1000"]}
    )
    assert len(should_not_exist_versions) == 0

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.integration
@pytest.mark.dependency()
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_sklearn(
    integration_test_url, project_name, deployment_mode, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"sklearn-sample-{deployment_mode_suffix(deployment_mode)}", ModelType.SKLEARN
    )

    model_dir = "test/sklearn-model"

    undeploy_all_version()

    with merlin.new_model_version() as v:
        merlin.log_model(model_dir=model_dir)

    endpoint = merlin.deploy(v, deployment_mode=deployment_mode)
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    merlin.undeploy(v)

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.integration
@pytest.mark.dependency()
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_xgboost(
    integration_test_url, project_name, deployment_mode, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"xgboost-sample-{deployment_mode_suffix(deployment_mode)}", ModelType.XGBOOST
    )

    model_dir = "test/xgboost-model"

    undeploy_all_version()

    with merlin.new_model_version() as v:
        # Upload the serialized model to MLP
        merlin.log_model(model_dir=model_dir)

    endpoint = merlin.deploy(v, deployment_mode=deployment_mode)
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    merlin.undeploy(v)

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.integration
@pytest.mark.dependency()
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_model_schema(
    integration_test_url, project_name, deployment_mode, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"model-schema-{deployment_mode_suffix(deployment_mode)}", ModelType.XGBOOST
    )

    model_dir = "test/xgboost-model"

    undeploy_all_version()
    model_schema = ModelSchema(spec=InferenceSchema(
        feature_types={
            "featureA": ValueType.FLOAT64,
            "featureB": ValueType.INT64,
            "featureC": ValueType.STRING,
            "featureD": ValueType.BOOLEAN
        },
        model_prediction_output=BinaryClassificationOutput(
            prediction_score_column="score",
            actual_score_column="actual",
            positive_class_label="completed",
            negative_class_label="non_complete",
            score_threshold=0.7
        )
    ))

    with merlin.new_model_version(model_schema=model_schema) as v:
        # Upload the serialized model to MLP
        merlin.log_model(model_dir=model_dir)

    assert v.model_schema.spec == model_schema.spec
    
    endpoint = merlin.deploy(v, deployment_mode=deployment_mode)
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    merlin.undeploy(v)

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.integration
def test_mlflow_tracking(
    integration_test_url, project_name, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("mlflow-test", ModelType.XGBOOST)

    model_dir = "test/xgboost-model"

    undeploy_all_version()

    with merlin.new_model_version() as v:
        merlin.log_model(model_dir=model_dir)
        merlin.log_param("model_type", "xgboost")
        merlin.log_param("iteration", 5)

        merlin.set_tag("version", "v1.0")
        merlin.set_tag("build", "latest")
        merlin.set_tag("team_id", 1)

        merlin.log_metric("model_loaded", 10.23)

        assert merlin.get_param("model_type") == "xgboost"
        # Stringify value which is integer originally
        assert merlin.get_param("iteration") == "5"
        assert merlin.get_param("random_key") is None

        assert merlin.get_tag("version") == "v1.0"
        assert merlin.get_tag("xxx") is None
        # Stringify value which is integer originally
        assert merlin.get_tag("team_id") == "1"

        assert merlin.get_metric("model_loaded") == 10.23
        assert merlin.get_metric("response_time") is None

        assert merlin.list_tag() == {
            "version": "v1.0",
            "build": "latest",
            "team_id": "1",
        }

        # TODO: Support downloading artifacts from S3 or S3-compatible alternative (such as MinIO)
        # merlin.download_artifact("test/downloaded_artifact")
        # artifact_dir = os.listdir('test/downloaded_artifact')
        # assert len(artifact_dir) > 0  # not empty directory

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.integration
@pytest.mark.dependency()
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_tensorflow(
    integration_test_url, project_name, deployment_mode, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"tensorflow-sample-{deployment_mode_suffix(deployment_mode)}",
        ModelType.TENSORFLOW,
    )

    model_dir = "test/tensorflow-model"

    undeploy_all_version()

    with merlin.new_model_version() as v:
        merlin.log_model(model_dir=model_dir)

    endpoint = merlin.deploy(v, deployment_mode=deployment_mode)
    resp = requests.post(f"{endpoint.url}", json=tensorflow_request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(tensorflow_request_json["instances"])

    merlin.undeploy(v)

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.pytorch
@pytest.mark.integration
@pytest.mark.dependency()
def test_pytorch(integration_test_url, project_name, use_google_oauth, requests):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pytorch-sample", ModelType.PYTORCH)

    model_dir = "test/pytorch-model/pytorch-sample"

    undeploy_all_version()

    with merlin.new_model_version() as v:
        merlin.log_model(model_dir=model_dir)
        endpoint = merlin.deploy(v)

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    merlin.undeploy(v)

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.serving
@pytest.mark.integration
def test_set_traffic(integration_test_url, project_name, use_google_oauth, requests):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("set-traffic-sample", ModelType.SKLEARN)

    model_dir = "test/sklearn-model"

    undeploy_all_version()

    with merlin.new_model_version() as v:
        # Upload the serialized model to MLP
        merlin.log_model(model_dir=model_dir)
        endpoint = merlin.deploy(v)

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    endpoint = merlin.set_traffic({v: 100})
    sleep(3)

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    # Try to undeploy serving model version. It must be fail
    with pytest.raises(Exception):
        assert merlin.undeploy(v)

    merlin.stop_serving_traffic(endpoint.environment_name)

    # Undeploy other running model version endpoints
    undeploy_all_version()

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.serving
@pytest.mark.integration
def test_serve_traffic(integration_test_url, project_name, use_google_oauth, requests):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("serve-traffic-sample", ModelType.SKLEARN)

    model_dir = "test/sklearn-model"

    undeploy_all_version()

    with merlin.new_model_version() as v:
        # Upload the serialized model to MLP
        merlin.log_model(model_dir=model_dir)
        endpoint = merlin.deploy(v)

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    model_endpoint = merlin.serve_traffic({endpoint: 100})
    sleep(3)

    resp = requests.post(f"{model_endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    # Try to undeploy serving model version. It must be fail
    with pytest.raises(Exception):
        assert merlin.undeploy(v)

    merlin.stop_serving_traffic(model_endpoint.environment_name)

    # Undeploy other running model version endpoints
    undeploy_all_version()

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.integration
def test_multi_env(integration_test_url, project_name, use_google_oauth, requests):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("multi-env", ModelType.XGBOOST)

    model_dir = "test/xgboost-model"

    envs = merlin.list_environment()
    assert len(envs) >= 1

    default_env = merlin.get_default_environment()
    assert default_env is not None

    undeploy_all_version()

    with merlin.new_model_version() as v:
        # Upload the serialized model to MLP
        merlin.log_model(model_dir=model_dir)
        endpoint = merlin.deploy(v, environment_name=default_env.name)

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    merlin.undeploy(v)

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.integration
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_resource_request(
    integration_test_url, project_name, deployment_mode, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"resource-request-{deployment_mode_suffix(deployment_mode)}", ModelType.XGBOOST
    )

    model_dir = "test/xgboost-model"

    envs = merlin.list_environment()
    assert len(envs) >= 1

    default_env = merlin.get_default_environment()
    assert default_env is not None

    undeploy_all_version()
    with merlin.new_model_version() as v:
        # Upload the serialized model to MLP
        merlin.log_model(model_dir=model_dir)

        resource_request = ResourceRequest(1, 1, "100m", "2", "256Mi")
        endpoint = merlin.deploy(
            v,
            environment_name=default_env.name,
            resource_request=resource_request,
            deployment_mode=deployment_mode,
        )

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    merlin.undeploy(v)

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.gpu
@pytest.mark.integration
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_resource_request_with_gpu(
    integration_test_url,
    project_name,
    deployment_mode,
    use_google_oauth,
    requests,
    gpu_config,
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"gpu-{deployment_mode_suffix(deployment_mode)}", ModelType.XGBOOST
    )

    model_dir = "test/xgboost-model"

    envs = merlin.list_environment()
    assert len(envs) >= 1

    default_env = merlin.get_default_environment()
    assert default_env is not None

    undeploy_all_version()
    with merlin.new_model_version() as v:
        # Upload the serialized model to MLP
        merlin.log_model(model_dir=model_dir)

        resource_request = ResourceRequest(1, 1, "100m", "2", "256Mi", **gpu_config)
        endpoint = merlin.deploy(
            v,
            environment_name=default_env.name,
            resource_request=resource_request,
            deployment_mode=deployment_mode,
        )

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    merlin.undeploy(v)


# https://github.com/kserve/kserve/issues/2142
# Logging is not supported yet in raw_deployment
@pytest.mark.integration
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_logger(
    integration_test_url, project_name, deployment_mode, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"logger-{deployment_mode_suffix(deployment_mode)}", ModelType.TENSORFLOW
    )
    model_dir = "test/tensorflow-model"

    undeploy_all_version()

    logger = Logger(model=LoggerConfig(enabled=True, mode=LoggerMode.ALL))
    with merlin.new_model_version() as v:
        merlin.log_model(model_dir=model_dir)
        endpoint = merlin.deploy(logger=logger, deployment_mode=deployment_mode)

    request_json = {
        "signature_name": "predict",
        "instances": [
            {
                "sepal_length": 2.8,
                "sepal_width": 1.0,
                "petal_length": 6.8,
                "petal_width": 0.4,
            },
            {
                "sepal_length": 0.1,
                "sepal_width": 0.5,
                "petal_length": 1.8,
                "petal_width": 2.4,
            },
        ],
    }
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(tensorflow_request_json["instances"])

    merlin.undeploy(v)

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.customtransformer
@pytest.mark.integration
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_custom_transformer(
    integration_test_url, project_name, deployment_mode, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"custom-transformer-{deployment_mode_suffix(deployment_mode)}",
        ModelType.CUSTOM,
    )

    undeploy_all_version()

    transformer = Transformer(
        "gcr.io/kubeflow-ci/kfserving/image-transformer:latest",
    )

    logger = Logger(
        model=LoggerConfig(enabled=True, mode=LoggerMode.ALL),
        transformer=LoggerConfig(enabled=True, mode=LoggerMode.ALL),
    )

    with merlin.new_model_version() as v:
        v.log_custom_model(image="ealen/echo-server:0.5.1", args="--port 8080")
        endpoint = merlin.deploy(
            transformer=transformer, logger=logger, deployment_mode=deployment_mode
        )

    assert endpoint.logger is not None

    model_config = endpoint.logger.model
    assert model_config is not None
    assert model_config.enabled
    assert model_config.mode == LoggerMode.ALL

    transformer_config = endpoint.logger.transformer
    assert transformer_config is not None
    assert transformer_config.enabled
    assert transformer_config.mode == LoggerMode.ALL

    with open(os.path.join("test/transformer", "input.json"), "r") as f:
        req = json.load(f)

    resp = requests.post(f"{endpoint.url}", json=req)

    assert resp.status_code == 200
    assert resp.json() is not None

    # Undeploy other running model version endpoints
    undeploy_all_version()


@pytest.mark.feast
@pytest.mark.integration
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_feast_enricher(
    integration_test_url, project_name, deployment_mode, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"feast-enricher-{deployment_mode_suffix(deployment_mode)}", ModelType.CUSTOM
    )

    undeploy_all_version()
    with merlin.new_model_version() as v:
        v.log_custom_model(image="ealen/echo-server:0.5.1", args="--port 8080")

    transformer_config_path = os.path.join("test/transformer", "feast_enricher.yaml")
    transformer = StandardTransformer(config_file=transformer_config_path, enabled=True)

    request_json = {"driver_id": "1000"}
    endpoint = merlin.deploy(
        v, transformer=transformer, deployment_mode=deployment_mode
    )
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    feast_features = resp.json()["request"]["body"]["feast_features"]
    assert feast_features is not None
    assert pd.DataFrame(feast_features) is not None

    merlin.undeploy(v)


@pytest.mark.integration
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_standard_transformer_without_feast(
    integration_test_url, project_name, deployment_mode, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"std-transformer-{deployment_mode_suffix(deployment_mode)}", ModelType.CUSTOM
    )

    undeploy_all_version()
    with merlin.new_model_version() as v:
        v.log_custom_model(image="ealen/echo-server:0.5.1", args="--port 8080")

    transformer_config_path = os.path.join(
        "test/transformer", "standard_transformer_no_feast.yaml"
    )
    transformer = StandardTransformer(
        config_file=transformer_config_path,
        enabled=True,
        env_vars={"MODEL_TIMEOUT": "5s"},
    )

    endpoint = merlin.deploy(
        v, transformer=transformer, deployment_mode=deployment_mode
    )
    request_json = {
        "drivers": [
            # 1 Feb 2022, 00:00:00
            {
                "id": 1,
                "name": "driver-1",
                "vehicle": "motorcycle",
                "previous_vehicle": "suv",
                "rating": 4,
                "ep_time": 1643673600,
            },
            # 30 Jan 2022, 00:00:00
            {
                "id": 2,
                "name": "driver-2",
                "vehicle": "sedan",
                "previous_vehicle": "mpv",
                "rating": 3,
                "ep_time": 1643500800,
            },
        ],
        "customer": {"id": 1111},
    }
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    exp_resp = {
        "instances": {
            "columns": [
                "customer_id",
                "name",
                "rank",
                "rating",
                "vehicle",
                "previous_vehicle",
                "ep_time_x",
                "ep_time_y",
            ],
            "data": [
                [1111, "driver-2", 2.5, 0.5, 2, 3, 1, 0],
                [1111, "driver-1", -2.5, 0.75, 0, 1, 1, 0],
            ],
        }
    }

    transformed_req = resp.json()["request"]["body"]
    recursive_eq(
        transformed_req["instances"], exp_resp["instances"], abs_tol=1e-09
    )  # asserts lhs = rhs, with tolerance
    merlin.undeploy(v)

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.feast
@pytest.mark.integration
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_standard_transformer_with_feast(
    integration_test_url, project_name, deployment_mode, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"std-transformer-feast-{deployment_mode_suffix(deployment_mode)}",
        ModelType.CUSTOM,
    )

    undeploy_all_version()
    with merlin.new_model_version() as v:
        v.log_custom_model(image="ealen/echo-server:0.5.1", args="--port 8080")

    transformer_config_path = os.path.join(
        "test/transformer", "standard_transformer_with_feast.yaml"
    )
    transformer = StandardTransformer(config_file=transformer_config_path, enabled=True)

    endpoint = merlin.deploy(
        v, transformer=transformer, deployment_mode=deployment_mode
    )
    request_json = {
        "drivers": [
            {"id": "1234", "name": "driver-1"},
            {"id": "5678", "name": "driver-2"},
        ],
        "customer": {"id": 1111},
    }
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    exp_resp = {
        "instances": {
            "columns": [
                "rank",
                "driver_id",
                "customer_id",
                "merlin_test_driver_features:test_int32",
                "merlin_test_driver_features:test_float",
            ],
            "data": [[0, "1234", 1111, -1, 0], [1, "5678", 1111, -1, 0]],
        }
    }

    assert resp.json()["request"]["body"]["instances"] == exp_resp["instances"]
    merlin.undeploy(v)


@pytest.mark.feast
@pytest.mark.integration
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_standard_transformer_with_multiple_feast(
    integration_test_url,
    project_name,
    deployment_mode,
    use_google_oauth,
    feast_serving_redis_url,
    feast_serving_bigtable_url,
    requests,
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"std-transformer-feasts-{deployment_mode_suffix(deployment_mode)}",
        ModelType.CUSTOM,
    )

    undeploy_all_version()
    with merlin.new_model_version() as v:
        v.log_custom_model(image="ealen/echo-server:0.5.1", args="--port 8080")

    config_template_file_path = os.path.join(
        "test/transformer", "standard_transformer_multiple_feast.yaml.tmpl"
    )
    config_file_path = os.path.join(
        "test/transformer", "standard_transformer_multiple_feast.yaml"
    )

    from string import Template

    config_template_file = open(config_template_file_path, "rt")
    t = Template(config_template_file.read())
    rendered_config = t.substitute(
        {
            "feast_serving_redis_url": feast_serving_redis_url,
            "feast_serving_bigtable_url": feast_serving_bigtable_url,
        }
    )
    config_file = open(config_file_path, "wt")
    config_file.write(rendered_config)
    config_file.close()

    transformer = StandardTransformer(config_file=config_file_path, enabled=True)

    endpoint = merlin.deploy(
        v, transformer=transformer, deployment_mode=deployment_mode
    )
    request_json = {
        "drivers": [
            {"id": "driver_1", "name": "driver-1"},
            {"id": "driver_2", "name": "driver-2"},
        ],
        "customer": {"id": 1111},
    }
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    exp_resp = {
        "instances": {
            "columns": [
                "rank",
                "driver_id",
                "customer_id",
                "merlin_test_redis_driver_features:completion_rate",
                "merlin_test_redis_driver_features:cancellation_rate",
                "merlin_test_bt_driver_features:rating",
            ],
            "data": [
                [0, "driver_1", 1111, 0.85, 0.15, 4.2],
                [1, "driver_2", 1111, 0.6, 0.4, 4.2],
            ],
        }
    }

    assert resp.json()["request"]["body"]["instances"] == exp_resp["instances"]
    merlin.undeploy(v)


@pytest.mark.skip(reason="Direct retrieval requires rework")
@pytest.mark.feast
@pytest.mark.integration
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_standard_transformer_with_multiple_feast_with_source(
    integration_test_url,
    project_name,
    deployment_mode,
    use_google_oauth,
    feast_serving_bigtable_url,
    requests,
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"std-trf-feasts-source-{deployment_mode_suffix(deployment_mode)}",
        ModelType.CUSTOM,
    )

    undeploy_all_version()
    with merlin.new_model_version() as v:
        v.log_custom_model(image="ealen/echo-server:0.5.1", args="--port 8080")

    config_template_file_path = os.path.join(
        "test/transformer", "standard_transformer_feast_with_source.yaml.tmpl"
    )
    config_file_path = os.path.join(
        "test/transformer", "standard_transformer_multiple_feast.yaml"
    )

    from string import Template

    config_template_file = open(config_template_file_path, "rt")
    t = Template(config_template_file.read())
    rendered_config = t.substitute(
        {
            "feast_serving_bigtable_url": feast_serving_bigtable_url,
        }
    )
    config_file = open(config_file_path, "wt")
    config_file.write(rendered_config)
    config_file.close()

    env_vars = {
        "FEAST_REDIS_DIRECT_STORAGE_ENABLED": True,
        "FEAST_REDIS_POOL_SIZE": 1,
        "FEAST_BIGTABLE_DIRECT_STORAGE_ENABLED": True,
        "FEAST_BIGTABLE_POOL_SIZE": 1,
        "FEAST_BIGTABLE_KEEP_ALIVE_INTERVAL": "2m",
        "FEAST_BIGTABLE_KEEP_ALIVE_TIMEOUT": "15s",
    }
    transformer = StandardTransformer(
        config_file=config_file_path, enabled=True, env_vars=env_vars
    )

    endpoint = merlin.deploy(
        v, transformer=transformer, deployment_mode=deployment_mode
    )
    request_json = {
        "drivers": [
            {"id": "driver_1", "name": "driver-1"},
            {"id": "driver_2", "name": "driver-2"},
        ],
        "customer": {"id": 1111},
    }
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    exp_resp = {
        "instances": {
            "columns": [
                "rank",
                "driver_id",
                "customer_id",
                "merlin_test_redis_driver_features:completion_rate",
                "merlin_test_redis_driver_features:cancellation_rate",
                "merlin_test_bt_driver_features:rating",
            ],
            "data": [
                [0, "driver_1", 1111, 0.85, 0.15, 4.2],
                [1, "driver_2", 1111, 0.6, 0.4, 4.2],
            ],
        }
    }

    assert resp.json()["request"]["body"]["instances"] == exp_resp["instances"]
    merlin.undeploy(v)


@pytest.mark.integration
def test_custom_model_without_artifact(
    integration_test_url, project_name, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("custom-wo-artifact", ModelType.CUSTOM)

    undeploy_all_version()

    with merlin.new_model_version() as v:
        v.log_custom_model(image="ealen/echo-server:0.5.1", args="--port 8080")

    endpoint = merlin.deploy(v)

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    # Undeploy other running model version endpoints
    undeploy_all_version()

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.integration
@pytest.mark.parametrize(
    "deployment_mode", [DeploymentMode.RAW_DEPLOYMENT, DeploymentMode.SERVERLESS]
)
def test_custom_model_with_artifact(
    integration_test_url, project_name, deployment_mode, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model(
        f"custom-w-artifact-{deployment_mode_suffix(deployment_mode)}", ModelType.CUSTOM
    )
    undeploy_all_version()

    model_dir = "test/custom-model"
    BST_FILE = "model.bst"

    with merlin.new_model_version() as v:
        v.log_custom_model(
            image="ghcr.io/tiopramayudi/custom-predictor-go:v0.2", model_dir=model_dir
        )

    endpoint = merlin.deploy(
        v,
        env_vars={"MODEL_FILE_NAME": BST_FILE},
        deployment_mode=deployment_mode,
    )

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert resp.json()["predictions"] is not None
    # Undeploy other running model version endpoints
    undeploy_all_version()

    print(f"Deleting model version created...")
    try:
        v.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.raw_deployment
@pytest.mark.integration
def test_deployment_mode_for_serving_model(
    integration_test_url, project_name, use_google_oauth, requests
):
    """
    Validate that set traffic is working when switching from different deployment mode
    """
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("serve-raw-deployment", ModelType.TENSORFLOW)
    model_dir = "test/tensorflow-model"

    undeploy_all_version()

    # Upload new model version: v1
    with merlin.new_model_version() as v1:
        merlin.log_model(model_dir=model_dir)

    # Deploy using serverless with RPS autoscaling policy
    endpoint = merlin.deploy(
        v1,
        autoscaling_policy=merlin.AutoscalingPolicy(
            metrics_type=merlin.MetricsType.RPS, target_value=20
        ),
    )

    resp = requests.post(f"{endpoint.url}", json=tensorflow_request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(tensorflow_request_json["instances"])

    # Set v1 as serving model
    initial_model_endpoint = merlin.set_traffic({v1: 100})
    resp = requests.post(f"{initial_model_endpoint.url}", json=tensorflow_request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(tensorflow_request_json["instances"])

    # Test that user can't change deployment mode of a serving model
    with pytest.raises(Exception):
        endpoint = merlin.deploy(v1, deployment_mode=DeploymentMode.RAW_DEPLOYMENT)

    # Upload new model version: v2
    with merlin.new_model_version() as v2:
        merlin.log_model(model_dir=model_dir)

    # Deploy v2 using raw_deployment with CPU autoscaling policy
    new_endpoint = merlin.deploy(
        v2,
        deployment_mode=DeploymentMode.RAW_DEPLOYMENT,
        autoscaling_policy=merlin.AutoscalingPolicy(
            metrics_type=merlin.MetricsType.CPU_UTILIZATION, target_value=20
        ),
    )

    resp = requests.post(f"{new_endpoint.url}", json=tensorflow_request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(tensorflow_request_json["instances"])

    # Set v2 as serving model
    model_endpoint = merlin.set_traffic({v2: 100})
    assert model_endpoint.url == initial_model_endpoint.url

    resp = requests.post(f"{model_endpoint.url}", json=tensorflow_request_json)
    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(tensorflow_request_json["instances"])

    # Set v1 back as serving model
    model_endpoint = merlin.set_traffic({v1: 100})
    assert model_endpoint.url == initial_model_endpoint.url

    resp = requests.post(f"{model_endpoint.url}", json=tensorflow_request_json)
    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(tensorflow_request_json["instances"])

    merlin.stop_serving_traffic(model_endpoint.environment_name)
    undeploy_all_version()

    print(f"Deleting model version created...")
    try:
        v1.delete_model_version()
        v2.delete_model_version()
    except Exception as e:
        print(e)
    print(f"Model version deleted.")


@pytest.mark.integration
def test_redeploy_model(integration_test_url, project_name, use_google_oauth, requests):
    """
    Validate that calling the 'merlin.deploy' twice in a row redeploys a Merlin model
    """

    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("model-sdk-redeploy", ModelType.CUSTOM)

    undeploy_all_version()

    # Upload new model version: v1
    with merlin.new_model_version() as v1:
        v1.log_custom_model(image="ealen/echo-server:0.5.1", args="--port 8080")

    # Deploy using raw_deployment with CPU autoscaling policy
    endpoint = merlin.deploy(
        v1,
        resource_request=merlin.ResourceRequest(1, 1, "123m", "2", "234Mi"),
        env_vars={"green": "TRUE"},
        autoscaling_policy=merlin.AutoscalingPolicy(
            metrics_type=merlin.MetricsType.CPU_UTILIZATION, target_value=50
        ),
        deployment_mode=DeploymentMode.RAW_DEPLOYMENT,
        transformer=Transformer(image="gcr.io/kubeflow-ci/kfserving/image-transformer:latest"),
    )

    with open(os.path.join("test/transformer", "input.json"), "r") as f:
        req = json.load(f)
    
    resp = requests.post(f"{endpoint.url}", json=req)
    assert resp.status_code == 200
    assert resp.json() is not None

    # Check the deployment configs of v1
    assert endpoint.resource_request.cpu_request == "123m"
    assert endpoint.resource_request.cpu_limit == "2"
    assert endpoint.resource_request.memory_request == "234Mi"
    assert endpoint.env_vars == {"green": "TRUE"}
    assert endpoint.deployment_mode == DeploymentMode.RAW_DEPLOYMENT

    # Check the autoscaling policy of v1
    assert endpoint.autoscaling_policy.metrics_type == MetricsType.CPU_UTILIZATION
    assert endpoint.autoscaling_policy.target_value == 50

    # Check that the transformer id is set
    assert endpoint.transformer.id != ""

    sleep(10)

    # Deploy v2 using raw_deployment with CPU autoscaling policy; update transformer specs.
    new_endpoint = merlin.deploy(
        v1,
        autoscaling_policy=merlin.AutoscalingPolicy(
            metrics_type=merlin.MetricsType.CPU_UTILIZATION, target_value=90
        ),
        deployment_mode=DeploymentMode.RAW_DEPLOYMENT,
        transformer=Transformer(image="gcr.io/kubeflow-ci/kfserving/image-transformer:latest",
                                resource_request=merlin.ResourceRequest(0, 1, "100m", "2", "250Mi")),
    )

    resp = requests.post(f"{new_endpoint.url}", json=req)
    assert resp.status_code == 200
    assert resp.json() is not None

    # Check that the endpoint remains the same
    assert endpoint.url == new_endpoint.url

    # Check that the deployment configs of v2 have remained the same as those used in v1
    assert endpoint.resource_request.cpu_request == "123m"
    assert endpoint.resource_request.cpu_limit == "2"
    assert endpoint.resource_request.memory_request == "234Mi"
    assert endpoint.env_vars == {"green": "TRUE"}
    assert endpoint.deployment_mode == DeploymentMode.RAW_DEPLOYMENT

    # Check the autoscaling policy of v2
    assert new_endpoint.autoscaling_policy.metrics_type == MetricsType.CPU_UTILIZATION
    assert new_endpoint.autoscaling_policy.target_value == 90

    # Check that the transformer id remains the same
    assert endpoint.transformer.id == new_endpoint.transformer.id

    undeploy_all_version()

    print(f"Deleting model versions created...")
    for v in merlin.active_model().list_version():
        try:
            v.delete_model_version()
        except Exception as e:
            print(e)
    print(f"Model versions deleted.")


def deployment_mode_suffix(deployment_mode: DeploymentMode):
    return deployment_mode.value.lower()[0:1]


@pytest.mark.integration
def test_standard_transformer_simulate(integration_test_url, use_google_oauth):
    """
    Test the `simulate` method of the `StandardTransformer` class.
    """
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)

    transformer_config_path = os.path.join(
        "test/transformer", "standard_transformer_no_feast.yaml"
    )
    transformer = StandardTransformer(
        config_file=transformer_config_path, enabled=False
    )

    payload = {
        "drivers": [
            # 1 Feb 2022, 00:00:00
            {
                "id": 1,
                "name": "driver-1",
                "vehicle": "motorcycle",
                "previous_vehicle": "suv",
                "rating": 4,
                "ep_time": 1643673600,
            },
            # 30 Jan 2022, 00:00:00
            {
                "id": 2,
                "name": "driver-2",
                "vehicle": "sedan",
                "previous_vehicle": "mpv",
                "rating": 3,
                "ep_time": 1643500800,
            },
        ],
        "customer": {"id": 1111},
    }

    resp_wo_tracing = transformer.simulate(payload=payload, exclude_tracing=True)
    resp_w_tracing = transformer.simulate(payload=payload, exclude_tracing=False)

    def remove_nulls(d):
        return {k: v for k, v in d.items() if v is not None}

    with open("test/transformer/sim_exp_resp_valid_wo_tracing.json", "r") as f:
        exp_resp_valid_wo_tracing = json.load(f, object_hook=remove_nulls)

    with open("test/transformer/sim_exp_resp_valid_w_tracing.json", "r") as f:
        exp_resp_valid_w_tracing = json.load(f, object_hook=remove_nulls)

    assert isinstance(resp_wo_tracing, dict)
    assert isinstance(resp_w_tracing, dict)
    assert "response" in resp_wo_tracing.keys()
    assert "response" in resp_w_tracing.keys()
    assert "operation_tracing" not in resp_wo_tracing.keys()
    assert "operation_tracing" in resp_w_tracing.keys()
    assert resp_wo_tracing == exp_resp_valid_wo_tracing
    assert resp_w_tracing == exp_resp_valid_w_tracing


@pytest.mark.feast
@pytest.mark.integration
def test_standard_transformer_simulate_feast(integration_test_url, use_google_oauth):
    """
    Test the `simulate` method of the `StandardTransformer` class with Feast enricher.
    """
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)

    transformer_config_path = os.path.join(
        "test/transformer", "standard_transformer_with_feast.yaml"
    )
    transformer = StandardTransformer(config_file=transformer_config_path, enabled=True)

    request_json = {
        "drivers": [
            {"id": "1234", "name": "driver-1"},
            {"id": "5678", "name": "driver-2"},
        ],
        "customer": {"id": 1111},
    }

    response_w_tracing = transformer.simulate(
        payload=request_json, exclude_tracing=True
    )
    response_wo_tracing = transformer.simulate(
        payload=request_json, exclude_tracing=False
    )

    with open("test/transformer/sim_exp_resp_feast_w_tracing.json", "r") as f:
        exp_resp_w_tracing = json.load(f)

    with open("test/transformer/sim_exp_resp_feast_wo_tracing.json", "r") as f:
        exp_resp_wo_tracing = json.load(f)

    assert isinstance(response_w_tracing, dict)
    assert isinstance(response_wo_tracing, dict)
    assert "response" in response_w_tracing.keys()
    assert "response" in response_wo_tracing.keys()
    assert "operation_tracing" not in response_wo_tracing.keys()
    assert "operation_tracing" in response_w_tracing.keys()
    assert response_w_tracing == exp_resp_w_tracing
    assert response_wo_tracing == exp_resp_wo_tracing
